@@ -1,5 +1,8 @@
-#include <SWI-cpp.h>
 #include "Rcpp.h"
+
+#include <SWI-cpp2.h>
+#include <SWI-cpp2.cpp>
+
 using namespace Rcpp ;
 
 // Translate prolog expression to R
@@ -44,51 +47,32 @@ RObject pl2r(PlTerm pl, CharacterVector& names, PlTerm& vars, List options) ;
 //
 PlTerm r2pl(SEXP r, CharacterVector& names, PlTerm& vars, List options) ;
 
-// Consult one or more files. If something fails, the procedure stops, and
-// will not try to consult the remaining files.
-//
-// [[Rcpp::export(.consult)]]
-LogicalVector consult_(CharacterVector files)
-{
-  for(R_xlen_t i=0; i<files.size(); i++)
-  {
-    try 
-    {
-      PlCall("consult", PlTermv(PlString(files(i)))) ;
-    }
-    
-    catch(PlException& ex)
-    { 
-      char* err = (char*) ex ;
-      PL_clear_exception() ;
-      stop("failed to consult %s: %s", (char*) files(i), err) ;
-    }
-  }
-
-  return true ;
-}
-
 // Prolog -> R
 RObject pl2r_null()
 {
   return R_NilValue ;
 }
 
+// TODO: use this and pl.eq_if_atom(ATOM_na) instead of
+//       pl.is_atom() && pl == "na"
+// PlAtom ATOM_na("na");
+
 // This helper function checks for na and then translates an individual PlTerm 
 // to a double.
 double pl2r_double(PlTerm pl)
 {
-  if(PL_is_atom(pl) && pl == "na")
+  if(pl.is_atom() && pl.as_string() == "na")
     return NA_REAL ;
 
   try 
   {
-    return (double) pl ;
+    return pl.as_float() ;
   }
 
   catch(PlException& ex)
   { 
-    warning("cannot convert %s to float: %s", (char*) pl, (char*) ex) ;
+    warning("cannot convert %s to float: %s",
+            pl.as_string(PlEncoding::Locale).c_str(), ex.as_string(PlEncoding::Locale).c_str()) ;
     PL_clear_exception() ;
     return NA_REAL ;
   }
@@ -103,8 +87,9 @@ DoubleVector pl2r_real(PlTerm pl)
 // Convert vector of reals (e.g., #(1.0, 2.0, na)) to DoubleVector
 DoubleVector pl2r_realvec(PlTerm pl)
 {
-  DoubleVector r(pl.arity()) ;
-  for(size_t i=0; i<pl.arity(); i++)
+  size_t arity = pl.arity() ;
+  DoubleVector r(arity) ;
+  for(size_t i=0; i<arity; i++)
     r(i) = pl2r_double(pl[i+1]) ;
 
   return r ;
@@ -117,12 +102,12 @@ NumericMatrix pl2r_realmat(PlTerm pl)
   size_t ncol = 0 ;
   if(nrow > 0)
   {
-    for(size_t i=0; i<pl.arity(); i++)
+    for(size_t i=0; i<nrow; i++)
       if(i == 0)
         ncol = pl[1].arity() ;
       else
       {
-        if(pl[i+1].arity() != ncol)
+        if(ncol != pl[i+1].arity())
           stop("cannot convert PlTerm to Matrix, inconsistent rows") ;
       }
   }
@@ -137,17 +122,18 @@ NumericMatrix pl2r_realmat(PlTerm pl)
 // See above for pl2r_double
 long pl2r_int(PlTerm pl)
 {
-  if(PL_is_atom(pl) && pl == "na")
+  if(pl.is_atom() && pl.as_string() == "na")
     return NA_INTEGER ;
 
   try 
   {
-    return (long) pl ;
+    return pl.as_long() ;
   }
   
   catch(PlException& ex)
   { 
-    warning("Cannot convert %s to integer: %s", (char*) pl, (char*) ex) ;
+    warning("Cannot convert %s to integer: %s",
+            pl.as_string(PlEncoding::Locale).c_str(), ex.as_string(PlEncoding::Locale).c_str()) ;
     PL_clear_exception() ;
     return NA_INTEGER ;
   }
@@ -160,8 +146,9 @@ IntegerVector pl2r_integer(PlTerm pl)
 
 IntegerVector pl2r_intvec(PlTerm pl)
 {
-  IntegerVector r(pl.arity()) ;
-  for(size_t i=0; i<pl.arity(); i++)
+  size_t arity = pl.arity() ;
+  IntegerVector r(arity) ;
+  for(size_t i=0; i<arity; i++)
     r(i) = pl2r_int(pl[i+1]) ;
 
   return r ;
@@ -173,7 +160,7 @@ IntegerMatrix pl2r_intmat(PlTerm pl)
   size_t ncol = 0 ;
   if(nrow > 0)
   {
-    for(size_t i=0; i<pl.arity(); i++)
+    for(size_t i=0; i<nrow; i++)
       if(i == 0)
         ncol = pl[1].arity() ;
       else
@@ -193,10 +180,10 @@ IntegerMatrix pl2r_intmat(PlTerm pl)
 // See above for pl2r_double
 String pl2r_string(PlTerm pl)
 {
-  if(PL_is_atom(pl) && pl == "na")
+  if(pl.is_atom() && pl.as_string() == "na")
     return NA_STRING ;
   
-  return (char*) pl ;
+  return pl.as_string(PlEncoding::Locale) ;
 }
 
 CharacterVector pl2r_char(PlTerm pl)
@@ -206,8 +193,9 @@ CharacterVector pl2r_char(PlTerm pl)
 
 CharacterVector pl2r_charvec(PlTerm pl)
 {
-  CharacterVector r(pl.arity()) ;
-  for(size_t i=0; i<pl.arity(); i++)
+  size_t arity = pl.arity() ;
+  CharacterVector r(arity) ;
+  for(size_t i=0; i<arity; i++)
     r(i) = pl2r_string(pl[i+1]) ;
 
   return r ;
@@ -219,9 +207,9 @@ CharacterMatrix pl2r_charmat(PlTerm pl)
   size_t ncol = 0 ;
   if(nrow > 0)
   {
-    for(size_t i=0; i<pl.arity(); i++)
+    for(size_t i=0; i<nrow; i++)
       if(i == 0)
-        ncol = pl[1].arity() ;
+        nrow = pl[1].arity() ;
       else
       {
         if(pl[i+1].arity() != ncol)
@@ -239,20 +227,20 @@ CharacterMatrix pl2r_charmat(PlTerm pl)
 // Convert prolog atom to R symbol (handle na, true, false)
 RObject pl2r_symbol(PlTerm pl)
 {
-  if(pl == "na")
+  if(pl.as_string() == "na")
     return wrap(NA_LOGICAL) ;
   
-  if(pl == "true")
+  if(pl.as_string() == "true")
     return wrap(true) ;
   
-  if(pl == "false")
+  if(pl.as_string() == "false")
     return wrap(false) ;
 
   // Empty symbols
-  if(pl == "")
+  if(pl.as_string() == "")
     return Function("substitute")() ;
 
-  return wrap(Symbol((char*) pl)) ;
+  return as<RObject>(Symbol(pl.as_string(PlEncoding::UTF8))) ; // TODO: PlEncoding::Locale?
 }
 
 // Forward declaration, needed below
@@ -264,62 +252,68 @@ RObject pl2r_function(PlTerm pl, CharacterVector& names, PlTerm& vars, List opti
   PlTerm plhead = pl[1] ;
   PlTerm plbody = pl[2] ;
 
-  Language head("alist") ;
-  for(unsigned int i=1 ; i<=plhead.arity() ; i++)
+  PlAtom n(PlAtom::null) ;
+  size_t arity = plhead.arity() ;
+
+  DottedPair head ;
+  for(unsigned int i=1 ; i<=arity ; i++)
   {
     PlTerm arg = plhead[i] ;
 
     // Compounds like mean=100 are translated to named function arguments
-    if(PL_is_compound(arg) && !strcmp(arg.name(), "=") && arg.arity() == 2)
+    if(arg.is_compound() && arg.name().as_string() == "=" && arg.arity() == 2)
     {
       PlTerm a1 = arg[1] ;
       PlTerm a2 = arg[2] ;
-      if(PL_is_atom(a1))
+      if(a1.is_atom())
       {
-        head.push_back(Named((char*) a1) = pl2r(a2, names, vars, options)) ;
+        head.push_back(Named(a1.as_string(PlEncoding::UTF8)) = pl2r(a2, names, vars, options)) ;
         continue ;
       }
     }
 
-    // just the name, no argument like in alist(a=, b=)
-    head.push_back(Named((char*) arg) = Function("substitute")()) ;
+    // the argument is the name
+    // head.push_back(Named(arg.as_string(PlEncoding::UTF8)) = pl2r_symbol(PlTerm_atom(""))) ;
+    head.push_back(Named(arg.as_string(PlEncoding::UTF8)) = Function("substitute")()) ;
   }
 
   RObject body = pl2r_compound(plbody, names, vars, options) ;
   head.push_back(body) ;
 
   Function as_function("as.function") ;
-  return wrap(as_function(head)) ;
+  Function as_list("as.list") ;
+  return wrap(as_function(as_list(head))) ;
 }
 
 LogicalVector pl2r_boolvec(PlTerm pl)
 {
-  LogicalVector r(pl.arity()) ;
-  for(size_t i=0; i<pl.arity(); i++)
+  size_t arity = pl.arity() ;
+  LogicalVector r(arity) ;
+  for(size_t i=0; i<arity; i++)
   {
     PlTerm t = pl[i+1] ;
-    if(PL_is_atom(t))
+    if(t.is_atom())
     {
-      if(t == "na")
+      if(t.as_string() == "na")
       {
         r(i) = NA_LOGICAL ;
         continue ;
       }
       
-      if(t == "true")
+      if(t.as_string() == "true")
       {
         r(i) = 1 ;
         continue ;
       }
       
-      if(t == "false")
+      if(t.as_string() == "false")
       {
         r(i) = 0 ;
         continue ;
       }
     }
     
-    warning("r2pl_logical: invalid item %s, returning NA", (char*) t) ;
+    warning("pl2r_logical: invalid item %s, returning NA", t.as_string(PlEncoding::Locale).c_str()) ;
     r(i) = NA_LOGICAL ;
   }
 
@@ -332,7 +326,7 @@ LogicalMatrix pl2r_boolmat(PlTerm pl)
   size_t ncol = 0 ;
   if(nrow > 0)
   {
-    for(size_t i=0; i<pl.arity(); i++)
+    for(size_t i=0; i<nrow; i++)
       if(i == 0)
         ncol = pl[1].arity() ;
       else
@@ -358,11 +352,11 @@ RObject pl2r_variable(PlTerm pl, CharacterVector& names, PlTerm& vars)
   //
   // Search for the variable (e.g., _1545) in names and return its R name as an
   // expression (say, X).
-  PlTail tail(vars) ;
-  PlTerm v ;
+  PlTerm_tail tail(vars) ;
+  PlTerm_var v ;
   for(int i=0 ; i<names.length() ; i++)
   {
-    tail.next(v) ;
+    PlCheckFail(tail.next(v)) ;
     if(v == pl)
       return ExpressionVector::create(Symbol(names(i))) ;
   }
@@ -370,7 +364,7 @@ RObject pl2r_variable(PlTerm pl, CharacterVector& names, PlTerm& vars)
   // If the variable is not found, it's a new one created by Prolog, e.g., in
   // queries like member(1, Y), Y is unified with [1 | _NewVar ]. This variable
   // cannot be translated to a human-readable name, so it is returned as _1545.
-  return ExpressionVector::create(Symbol((char*) pl)) ;
+  return ExpressionVector::create(Symbol(pl.as_string(PlEncoding::UTF8))) ; // TODO: PlEncoding::Locale?
 }
 
 // Translate prolog compound to R call
@@ -381,59 +375,59 @@ RObject pl2r_variable(PlTerm pl, CharacterVector& names, PlTerm& vars)
 RObject pl2r_compound(PlTerm pl, CharacterVector& names, PlTerm& vars, List options)
 {
   // This function does not (yet) work for cyclic terms
-  if(!PL_is_acyclic(pl))
-    stop("pl2r: Cannot convert cyclic term %s", (char*) pl) ;
+  if(!PL_is_acyclic(pl.C_))
+    stop("pl2r: Cannot convert cyclic term %s", pl.as_string(PlEncoding::Locale).c_str()) ;
 
   // Convert ##(#(...), ...) to NumericMatrix
-  if(!strcmp(pl.name(), options("realmat")))
+  if(!strcmp(pl.name().as_string(PlEncoding::UTF8).c_str(), options("realmat")))
     return pl2r_realmat(pl) ;
 
   // Convert #(1.0, 2.0, 3.0) to DoubleVector (# given by options("realvec"))
-  if(!strcmp(pl.name(), options("realvec")))
+  if(!strcmp(pl.name().as_string(PlEncoding::UTF8).c_str(), options("realvec")))
     return pl2r_realvec(pl) ;
 
   // Convert %%(%(...), ...) to IntegerMatrix
-  if(!strcmp(pl.name(), options("intmat")))
+  if(!strcmp(pl.name().as_string(PlEncoding::UTF8).c_str(), options("intmat")))
     return pl2r_intmat(pl) ;
 
   // Convert %(1.0, 2.0, 3.0) to IntegerVector
-  if(!strcmp(pl.name(), options("intvec")))
+  if(!strcmp(pl.name().as_string(PlEncoding::UTF8).c_str(), options("intvec")))
     return pl2r_intvec(pl) ;
 
   // Convert $$$($$(...), ...) to StringMatrix
-  if(!strcmp(pl.name(), options("charmat")))
+  if(!strcmp(pl.name().as_string(PlEncoding::UTF8).c_str(), options("charmat")))
     return pl2r_charmat(pl) ;
 
   // Convert $$(1.0, 2.0, 3.0) to CharacterVector
-  if(!strcmp(pl.name(), options("charvec")))
+  if(!strcmp(pl.name().as_string(PlEncoding::UTF8).c_str(), options("charvec")))
     return pl2r_charvec(pl) ;
 
   // Convert !!(!(...), ...) to LogicalMatrix
-  if(!strcmp(pl.name(), options("boolmat")))
+  if(!strcmp(pl.name().as_string(PlEncoding::UTF8).c_str(), options("boolmat")))
     return pl2r_boolmat(pl) ;
 
   // Convert !(1.0, 2.0, 3.0) to LogicalVector
-  if(!strcmp(pl.name(), options("boolvec")))
+  if(!strcmp(pl.name().as_string(PlEncoding::UTF8).c_str(), options("boolvec")))
     return pl2r_boolvec(pl) ;
 
   // Convert :- to function
-  if(!strcmp(pl.name(), ":-"))
+  if(pl.name().as_string() == ":-")
     return pl2r_function(pl, names, vars, options) ;
 
   // Other compounds
-  Language r(pl.name()) ;
+  Language r(pl.name().as_string(PlEncoding::UTF8).c_str()) ;
   for(unsigned int i=1 ; i<=pl.arity() ; i++)
   {
     PlTerm arg = pl[i] ;
 
     // Compounds like mean=100 are translated to named function arguments
-    if(PL_is_compound(arg) && !strcmp(arg.name(), "=") && arg.arity() == 2)
+    if(arg.is_compound() && !strcmp(arg.name().as_string(PlEncoding::UTF8).c_str(), "=") && arg.arity() == 2)
     {
       PlTerm a1 = arg[1] ;
       PlTerm a2 = arg[2] ;
-      if(PL_is_atom(a1))
+      if(a1.is_atom())
       {
-        r.push_back(Named(a1.name()) = pl2r(a2, names, vars, options)) ;
+        r.push_back(Named(a1.name().as_string(PlEncoding::UTF8).c_str()) = pl2r(a2, names, vars, options)) ;
         continue ;
       }
     }
@@ -467,13 +461,13 @@ RObject pl2r_list(PlTerm pl, CharacterVector& names, PlTerm& vars, List options)
     List r = as<List>(tail) ;
     
     // convert prolog pair a-X to named list element
-    if(PL_is_compound(head) && !strcmp(head.name(), "-") && head.arity() == 2)
+    if(head.is_compound() && !strcmp(head.name().as_string(PlEncoding::UTF8).c_str(), "-") && head.arity() == 2)
     {
       PlTerm a1 = head[1] ;
       PlTerm a2 = head[2] ;
-      if(PL_is_atom(a1))
+      if(a1.is_atom())
       {
-        r.push_front(pl2r(a2, names, vars, options), a1.name()) ;
+        r.push_front(pl2r(a2, names, vars, options), a1.name().as_string(PlEncoding::UTF8).c_str()) ;
         return r ;
       }
     }
@@ -484,16 +478,16 @@ RObject pl2r_list(PlTerm pl, CharacterVector& names, PlTerm& vars, List options)
   }
     
   // if the tail is something else, return [|](head, tail)
-  Language r(pl.name()) ;
+  Language r(pl.name().as_string(PlEncoding::UTF8).c_str()) ;
   
   // convert prolog pair a-X to named list element
-  if(PL_is_compound(head) && !strcmp(head.name(), "-") && head.arity() == 2)
+  if(head.is_compound() && !strcmp(head.name().as_string(PlEncoding::UTF8).c_str(), "-") && head.arity() == 2)
   {
     PlTerm a1 = head[1] ;
     PlTerm a2 = head[2] ;
-    if(PL_is_atom(a1))
+    if(a1.is_atom())
     {
-      r.push_back(Named(a1.name()) = pl2r(a2, names, vars, options)) ;
+      r.push_back(Named(a1.name().as_string(PlEncoding::UTF8).c_str()) = pl2r(a2, names, vars, options)) ;
       r.push_back(tail) ;
       return as<RObject>(r) ;
     }
@@ -509,30 +503,29 @@ RObject pl2r(PlTerm pl, CharacterVector& names, PlTerm& vars, List options)
 {
   if(pl.type() == PL_NIL)
     return pl2r_null() ;
-
-  if(PL_is_integer(pl))
+  
+  if(pl.is_integer())
     return pl2r_integer(pl) ;
-
-  if(PL_is_float(pl))
+  
+  if(pl.is_float())
     return pl2r_real(pl) ;
-
-  if(PL_is_string(pl))
+  
+  if(pl.is_string())
     return pl2r_char(pl) ;
-
-  if(PL_is_atom(pl))
+  
+  if(pl.is_atom())
     return pl2r_symbol(pl) ;
-
-  if(PL_is_list(pl))
+  
+  if(pl.is_list())
     return pl2r_list(pl, names, vars, options) ;
-
-  if(PL_is_compound(pl))
+  
+  if(pl.is_compound())
     return pl2r_compound(pl, names, vars, options) ;
-
-  if(PL_is_variable(pl))
+  
+  if(pl.is_variable())
     return pl2r_variable(pl, names, vars) ;
-
-  stop("pl2r: Cannot convert %s", (char*) pl) ;
-  return pl2r_null() ;
+  
+  stop("pl2r: Cannot convert %s", pl.as_string(PlEncoding::Locale).c_str()) ;
 }
 
 // Translate R expression to prolog
@@ -546,15 +539,16 @@ PlTerm r2pl_string(CharacterVector r, List options) ;
 // This returns an empty list
 PlTerm r2pl_null()
 {
-  PlTerm pl ;
-  PlTail(pl).close() ;
-  return pl ;
+  PlTerm_var pl ;
+  PlTerm_tail tail(pl) ;
+  PlCheckFail(tail.close()) ;
+  return tail ;
 }
 
 // Prolog representation of R's NA.
 PlTerm r2pl_na()
 {
-  return PlAtom("na") ;
+  return PlTerm_atom("na") ;
 }
 
 // Translate to matrix ###(##(1.0, 2.0, 3.0), ##(4.0, 5.0, 6.0))
@@ -565,15 +559,9 @@ PlTerm r2pl_matrix(Matrix<REALSXP> r, List aoptions)
 
   PlTermv rows(r.nrow()) ;
   for(int i=0 ; i<r.nrow() ; i++)
-    rows[i] = r2pl_real(r.row(i), options) ;
+    PlCheckFail(rows[i].unify_term(r2pl_real(r.row(i), options))) ;
 
-  PlTermv matrix(5) ;
-  matrix[0] = PlTerm((long int) r.nrow()) ;
-  matrix[1] = PlTerm((long int) r.ncol()) ;
-  matrix[2] = r2pl_string(rownames(r), options) ;
-  matrix[3] = r2pl_string(colnames(r), options) ;
-  matrix[4] = PlCompound("data", rows) ;
-  return PlCompound((const char*) aoptions("realmat"), matrix) ;
+  return PlCompound((const char*) options("realmat"), rows) ;
 }
 
 // Translate to (scalar) real or compounds like ##(1.0, 2.0, 3.0)
@@ -594,24 +582,24 @@ PlTerm r2pl_real(NumericVector r, List options)
     if(na[0] && !nan[0])
       return r2pl_na() ;
     
-    return PlTerm((double) r[0]);
+    return PlTerm_float(r[0]);
   }
 
-  // Translate to vector ##(1.0, 2.0, 3.0)
+  // Translate to vector #(1.0, 2.0, 3.0)
   size_t len = (size_t) r.length() ;
   PlTermv args(len) ;
   for(size_t i=0 ; i<len ; i++)
   {
     if(na[i] && !nan[i])
-      args[i] = r2pl_na() ;
+      PlCheckFail(args[i].unify_term(r2pl_na())) ;
     else
-      args[i] = PlTerm((double) r[i]) ;
+      PlCheckFail(args[i].unify_float(r[i])) ;
   }
 
   return PlCompound((const char*) options("realvec"), args) ;
 }
 
-// Translate to matrix !!(!(true, false), !(false, true))
+// Translate to matrix !!!(!!(true, false), !(false, true))
 PlTerm r2pl_matrix(Matrix<LGLSXP> r, List aoptions)
 {
   List options(aoptions) ;
@@ -619,12 +607,12 @@ PlTerm r2pl_matrix(Matrix<LGLSXP> r, List aoptions)
 
   PlTermv rows(r.nrow()) ;
   for(int i=0 ; i<r.nrow() ; i++)
-    rows[i] = r2pl_logical(r.row(i), options) ;
+    PlCheckFail(rows[i].unify_term(r2pl_logical(r.row(i), options))) ;
 
   return PlCompound((const char*) options("boolmat"), rows) ;
 }
 
-// Translate to (scalar) boolean or compounds like !(true, false, na)
+// Translate to (scalar) boolean or compounds like !!(true, false, na)
 PlTerm r2pl_logical(LogicalVector r, List options)
 {
   if(Rf_isMatrix(r))
@@ -641,7 +629,7 @@ PlTerm r2pl_logical(LogicalVector r, List options)
     if(na[0])
       return r2pl_na() ;
     
-    return PlAtom(r[0] ? "true" : "false") ;
+    return PlTerm_atom(r[0] ? "true" : "false") ;
   }
 
   // LogicalVector !(true, false, na)
@@ -650,15 +638,15 @@ PlTerm r2pl_logical(LogicalVector r, List options)
   for(size_t i=0 ; i<len ; i++)
   {
     if(na[i])
-      args[i] = r2pl_na() ;
+      PlCheckFail(args[i].unify_term(r2pl_na())) ;
     else
-      args[i] = PlAtom(r[i] ? "true" : "false") ;
+      PlCheckFail(args[i].unify_atom(r[i] ? "true" : "false")) ;  // TODO: unify_bool()
   }
 
   return PlCompound((const char*) options("boolvec"), args) ;
 }
 
-// Translate to matrix %%(%(1, 2), %(3, 4))
+// Translate to matrix %%%(%%(1, 2), %(3, 4))
 PlTerm r2pl_matrix(Matrix<INTSXP> r, List aoptions)
 {
   List options(aoptions) ;
@@ -666,12 +654,12 @@ PlTerm r2pl_matrix(Matrix<INTSXP> r, List aoptions)
 
   PlTermv rows(r.nrow()) ;
   for(int i=0 ; i<r.nrow() ; i++)
-    rows[i] = r2pl_integer(r.row(i), options) ;
+    PlCheckFail(rows[i].unify_term(r2pl_integer(r.row(i), options))) ;
 
   return PlCompound((const char*) options("intmat"), rows) ;
 }
 
-// Translate to (scalar) integer or compounds like %(1, 2, 3)
+// Translate to (scalar) integer or compounds like %%(1, 2, 3)
 PlTerm r2pl_integer(IntegerVector r, List options)
 {
   if(Rf_isMatrix(r))
@@ -688,7 +676,7 @@ PlTerm r2pl_integer(IntegerVector r, List options)
     if(na[0])
       return r2pl_na() ;
     
-    return PlTerm((long) r(0)) ;
+    return PlTerm_integer(r(0)) ;
   }
   
   // IntegerVector %(1, 2, 3)
@@ -697,9 +685,9 @@ PlTerm r2pl_integer(IntegerVector r, List options)
   for(size_t i=0 ; i<len ; i++)
   {
     if(na[i])
-      args[i] = r2pl_na() ;
+      PlCheckFail(args[i].unify_term(r2pl_na())) ;
     else
-      args[i] = PlTerm((long) r[i]) ;
+      PlCheckFail(args[i].unify_integer(r[i])) ;
   }
   
   return PlCompound((const char*) options("intvec"), args) ;
@@ -713,8 +701,6 @@ PlTerm r2pl_integer(IntegerVector r, List options)
 // variables, and it is unified with it if the name is found. Otherwise, a new
 // variable is created.
 //
-// If options("atomize") is true, no variable is created, but an atom is created 
-// with the variable name from R. This is only used for pretty printing.
 PlTerm r2pl_var(ExpressionVector r, CharacterVector& names, PlTerm& vars, List options)
 {
   // Variable name in R
@@ -722,33 +708,33 @@ PlTerm r2pl_var(ExpressionVector r, CharacterVector& names, PlTerm& vars, List o
   
   // If the variable should be "atomized" for pretty printing
   if(as<LogicalVector>(options("atomize"))(0))
-    return PlAtom(n.c_str()) ;
+    return PlTerm_atom(n.c_str()) ; // TODO: 
 
   // Do not map the anonymous variable to a known variable name
   if(n == "_")
-    return PlTerm() ;
+    return PlTerm_var() ;
 
   // Unify with existing variable of the same name
-  PlTail tail(vars) ;
-  PlTerm v ;
+  PlTerm_tail tail(vars) ;
+  PlTerm_var v ;
   for(R_xlen_t i=0 ; i<names.length() ; i++)
   {
-    tail.next(v) ;
+    PlCheckFail(tail.next(v)) ;
     if(n == names(i))
       return v ;
   }
 
   // If no such variable exists, create a new one and remember the name
   names.push_back(n.c_str()) ;
-  PlTerm pl ;
-  tail.append(pl) ;
+  PlTerm_var pl ;
+  PlCheckFail(tail.append(pl)) ;
   return pl ;
 }
 
 // Translate R symbol to prolog atom
 PlTerm r2pl_atom(Symbol r)
 {
-  return PlAtom(r.c_str()) ;
+  return PlTerm_atom(r.c_str()) ;
 }
 
 // Translate to matrix $$$($$(1, 2), $$(3, 4))
@@ -759,7 +745,7 @@ PlTerm r2pl_matrix(Matrix<STRSXP> r, List aoptions)
 
   PlTermv rows(r.nrow()) ;
   for(int i=0 ; i<r.nrow() ; i++)
-    rows[i] = r2pl_string(r.row(i), options) ;
+    PlCheckFail(rows[i].unify_term(r2pl_string(r.row(i), options))) ;
 
   return PlCompound((const char*) options("charmat"), rows) ;
 }
@@ -781,7 +767,7 @@ PlTerm r2pl_string(CharacterVector r, List options)
     if(na[0])
       return r2pl_na() ;
     
-    return PlString(r(0)) ;
+    return PlTerm_string(r(0)) ;
   }
 
   // compound like $("a", "b", "c")
@@ -790,9 +776,9 @@ PlTerm r2pl_string(CharacterVector r, List options)
   for(size_t i=0 ; i<len ; i++)
   {
     if(na[i])
-      args[i] = r2pl_na() ;
+      PlCheckFail(args[i].unify_term(r2pl_na())) ;
     else
-      args[i] = PlString(r(i)) ;
+      PlCheckFail(args[i].unify_term(PlTerm_string(r(i)))) ; // DO NOT SUBMIT - unify_string()
   }
 
   return PlCompound((const char*) options("charvec"), args) ;
@@ -810,8 +796,8 @@ PlTerm r2pl_compound(Language r, CharacterVector& names, PlTerm& vars, List opti
   if(len == 0)
   {
     PlTermv pl(3) ;
-    pl[1] = as<Symbol>(CAR(r)).c_str() ;
-    pl[2] = (long) 0 ;
+    PlCheckFail(pl[1].unify_atom(as<Symbol>(CAR(r)).c_str())) ;
+    PlCheckFail(pl[2].unify_integer(0)) ;
     PlCall("compound_name_arity", pl) ;
     return pl[0] ;
   }
@@ -829,9 +815,9 @@ PlTerm r2pl_compound(Language r, CharacterVector& names, PlTerm& vars, List opti
     
     // Convert named arguments to prolog compounds a=X
     if(n.length() && n(i) != "")
-      pl[i] = PlCompound("=", PlTermv(PlAtom(n(i)), arg)) ;
+      PlCheckFail(pl[i].unify_term(PlCompound("=", PlTermv(PlTerm_atom(n(i)), arg)))) ;
     else
-      pl[i] = arg ; // no name
+      PlCheckFail(pl[i].unify_term(arg)) ; // no name
   }
 
   return PlCompound(as<Symbol>(CAR(r)).c_str(), pl) ;
@@ -849,20 +835,20 @@ PlTerm r2pl_list(List r, CharacterVector& names, PlTerm& vars, List options)
   if(TYPEOF(r.names()) == STRSXP)
     n = as<CharacterVector>(r.names()) ;
   
-  PlTerm pl ;
-  PlTail tail(pl) ;
+  PlTerm_var pl ;
+  PlTerm_tail tail(pl) ;
   for(R_xlen_t i=0; i<r.size() ; i++)
   {
     PlTerm arg = r2pl(r(i), names, vars, options) ;
     
     // Convert named argument to prolog pair a-X.
     if(n.length() && n(i) != "")
-      tail.append(PlCompound("-", PlTermv(PlAtom(n(i)), arg))) ;
+      PlCheckFail(tail.append(PlCompound("-", PlTermv(PlTerm_atom(n(i)), arg)))) ;
     else
-      tail.append(arg) ; // no name
+      PlCheckFail(tail.append(arg)) ; // no name
   }
   
-  tail.close() ;
+  PlCheckFail(tail.close()) ;
   return pl ;
 }
 
@@ -870,55 +856,30 @@ PlTerm r2pl_list(List r, CharacterVector& names, PlTerm& vars, List options)
 PlTerm r2pl_function(Function r, CharacterVector& names, PlTerm& vars, List options)
 {
   PlTermv fun(2) ;
-  fun[1] = r2pl_compound(BODY(r), names, vars, options) ;
-  
+#if defined(R_VERSION) && R_VERSION >= R_Version(4, 5, 0)
+  PlCheckFail(fun[1].unify_term(r2pl_compound(R_ClosureBody(r), names, vars, options))) ;
+  List formals = as<List>(R_ClosureFormals(r)) ;
+#else
+  PlCheckFail(fun[1].unify_term(r2pl_compound(BODY(r), names, vars, options))) ;
   List formals = as<List>(FORMALS(r)) ;
+#endif
   size_t len = (size_t) formals.size() ;
   if(len == 0)
   {
     PlTermv pl(3) ;
-    pl[1] = PlAtom("function") ;
-    pl[2] = (long) 0 ;
+    PlCheckFail(pl[1].unify_atom("$function")) ;
+    PlCheckFail(pl[2].unify_integer(0)) ;
     PlCall("compound_name_arity", pl) ;
 
-    fun[0] = pl[0] ;
+    PlCheckFail(fun[0].unify_term(pl[0])) ;
     return PlCompound(":-", fun) ;
   }
   
   CharacterVector n = formals.names() ;
   PlTermv pl(len) ;
   for(size_t i=0 ; i<len ; i++)
-    pl[i] = PlAtom(n(i)) ;
-
-  fun[0] = PlCompound("function", pl) ;
-  return PlCompound(":-", fun) ;
-}
-
-// Translate R primitive to :- ("neck")
-PlTerm r2pl_builtin(Function r, CharacterVector& names, PlTerm& vars, List options)
-{
-  PlTermv fun(2) ;
-  fun[1] = r2pl_null() ;
-  
-  List formals = as<List>(FORMALS(r)) ;
-  size_t len = (size_t) formals.size() ;
-  if(len == 0)
-  {
-    PlTermv pl(3) ;
-    pl[1] = PlAtom("function") ;
-    pl[2] = (long) 0 ;
-    PlCall("compound_name_arity", pl) ;
-
-    fun[0] = pl[0] ;
-    return PlCompound(":-", fun) ;
-  }
-  
-  CharacterVector n = formals.names() ;
-  PlTermv pl(len) ;
-  for(size_t i=0 ; i<len ; i++)
-    pl[i] = PlAtom(n(i)) ;
-
-  fun[0] = PlCompound("function", pl) ;
+    PlCheckFail(pl[i].unify_atom(n(i))) ;
+  PlCheckFail(fun[0].unify_term(PlCompound("$function", pl))) ;
   return PlCompound(":-", fun) ;
 }
 
@@ -953,32 +914,29 @@ PlTerm r2pl(SEXP r, CharacterVector& names, PlTerm& vars, List options)
   
   if(TYPEOF(r) == CLOSXP)
     return r2pl_function(r, names, vars, options) ;
-
-  if(TYPEOF(r) == BUILTINSXP)
-    return r2pl_builtin(r, names, vars, options) ;
-
-  warning("r2pl: cannot translate R object of class %s, returning NA\n", 
-    as<CharacterVector>(Function("class")(r))) ;
+  
   return r2pl_na() ;
 }
 
-class RlQuery  
+#ifdef RPACKAGE
+
+class RlQuery
 {
   CharacterVector names ;
-  PlTerm vars ;
+  PlTerm_var vars ;
   List options ;
   Environment env ;
   PlQuery* qid ;
-  
+
 public:
   RlQuery(RObject aquery, List aoptions, Environment aenv) ;
   ~RlQuery() ;
-  
+
   int next_solution() ;
 
   List bindings() ;
 
-  const List& get_options() const 
+  const List& get_options() const
   {
     return options ;
   }
@@ -1011,7 +969,7 @@ int RlQuery::next_solution()
 {
   if(qid == NULL)
     stop("next_solution: no open query.") ;
-    
+
   int q ;
   try
   {
@@ -1020,24 +978,61 @@ int RlQuery::next_solution()
 
   catch(PlException& ex)
   {
-    char* err = (char*) ex ;
+    warning(ex.as_string(PlEncoding::Locale).c_str()) ;
     PL_clear_exception() ;
-    warning("Prolog exception: %s", err) ;
-    stop(err) ;
+    stop("Query failed") ;
   }
 
   return q ;
+
+/*
+  int q = qid->next_solution() ;
+  if(q == PL_S_TRUE)
+    return true ;
+
+  if(q == PL_S_LAST)
+    return true ;
+
+  if(q == PL_S_FALSE)
+  {
+    PL_close_query(qid) ;
+    qid = 0 ;
+    return false ;
+  }
+
+  if(q == PL_S_EXCEPTION)
+  {
+    PL_close_query(qid) ;
+    qid = 0 ;
+
+    term_t ex = PL_exception(0) ;
+    char* err ;
+    if(PL_get_chars(ex, &err, BUF_DISCARDABLE|CVT_WRITE|REP_UTF8))
+    {
+      PL_clear_exception() ;
+      warning(err) ;
+      return false ;
+    }
+
+    PL_clear_exception() ;
+    warning("query: unknown exception occurred") ;
+    return false ;
+  }
+
+  // should not be reached
+  return q ;
+  */
 }
 
 List RlQuery::bindings()
 {
   List l ;
 
-  PlTail tail(vars) ;
-  PlTerm v ;
+  PlTerm_tail tail(vars) ;
+  PlTerm_var v ;
   for(int i=0 ; i<names.length() ; i++)
   {
-    tail.next(v) ;
+    PlCheckFail(tail.next(v)) ;
     RObject r = pl2r(v, names, vars, options) ;
     if(TYPEOF(r) == EXPRSXP && names[i] == as<Symbol>(as<ExpressionVector>(r)[0]).c_str())
     continue ;
@@ -1048,13 +1043,13 @@ List RlQuery::bindings()
   return l ;
 }
 
-RlQuery* query_id = NULL ;
+static RlQuery* query_id = NULL ;
 
 // Open a query for later use.
 // [[Rcpp::export(.query)]]
 RObject query_(RObject query, List options, Environment env)
 {
-  if(PL_current_query() != 0)
+  if(query_id || PL_current_query())
   {
     warning("Cannot raise simultaneous queries. Please invoke clear()") ;
     return wrap(false) ;
@@ -1142,8 +1137,31 @@ List findall_(RObject query, List options, Environment env)
   return results ;
 }
 
-// Pretty print query. Maybe simplify to something like this:
-// with_output_to(string(S), write_term(member(X), [variable_names(['X'=X])])).
+// Consult one or more files. If something fails, the procedure stops, and
+// will not try to consult the remaining files.
+//
+// [[Rcpp::export(.consult)]]
+LogicalVector consult_(CharacterVector files)
+{
+  for(R_xlen_t i=0; i<files.size(); i++)
+  {
+    try 
+    {
+      PlCall("consult", PlTermv(PlTerm_string(files(i))));
+    }
+    
+    catch(PlException& ex)
+    { 
+      String err(ex.as_string(PlEncoding::Locale));
+      PL_clear_exception() ;
+      stop("failed to consult %s: %s", (char*) files(i), err.get_cstring()) ;
+    }
+  }
+
+  return true ;
+}
+
+// Pretty print query
 //
 // [[Rcpp::export(.portray)]]
 RObject portray_(RObject query, List options)
@@ -1155,14 +1173,14 @@ RObject portray_(RObject query, List options)
   }
 
   CharacterVector names ;
-  PlTerm vars ;
+  PlTerm_var vars ;
   options("atomize") = true ; // translate variables to their R names
   PlTermv pl(3) ;
-  pl[0] = r2pl(query, names, vars, options) ;
-  PlTail tail(pl[2]) ;
-  tail.append(PlCompound("quoted", PlTermv(PlAtom("false")))) ;
-  tail.append(PlCompound("spacing", PlTermv(PlAtom("next_argument")))) ;
-  tail.close() ;
+  PlCheckFail(pl[0].unify_term(r2pl(query, names, vars, options))) ;
+  PlTerm_tail tail(pl[2]) ;
+  PlCheckFail(tail.append(PlCompound("quoted", PlTermv(PlTerm_atom("false"))))) ;
+  PlCheckFail(tail.append(PlCompound("spacing", PlTermv(PlTerm_atom("next_argument"))))) ;
+  PlCheckFail(tail.close()) ;
 
   PlFrame f ;
   PlQuery q("term_string", pl) ;
@@ -1174,45 +1192,71 @@ RObject portray_(RObject query, List options)
   
   catch(PlException& ex)
   {
-    char* err = (char*) ex ;
+    warning(ex.as_string(PlEncoding::Locale).c_str()) ;
     PL_clear_exception() ;
-    stop("portray of %s failed: %s", (char*) pl[0], err) ;
+    stop("portray of %s failed.", pl[0].as_string(PlEncoding::Locale).c_str()) ;
   }
   
   return pl2r(pl[1], names, vars, options) ;
 }
 
+// Execute a query given as a string
+//
+// Example:
+// once("use_module(library(http/html_write))")
+//
+// [[Rcpp::export(.call)]]
+RObject call_(String query)
+{
+  if(PL_current_query() != 0)
+  {
+    warning("Closing the current query.") ;
+    clear_() ;
+  }
+
+  bool r = false ;
+  try
+  {
+    r = PlCall(query.get_cstring()) ;
+  }
+  
+  catch(PlException& ex)
+  {
+    warning(ex.as_string(PlEncoding::Locale).c_str()) ;
+    PL_clear_exception() ;
+    stop("query failed: %s", query.get_cstring()) ;
+  }
+  
+  return wrap(r) ;
+}
+
 // Call R expression from Prolog
 PREDICATE(r_eval, 1)
 {
   CharacterVector names ;
-  PlTerm vars ;
+  PlTerm_var vars ;
   List options ;
   if(query_id)
     options = query_id->get_options() ;
   else
-    options = List::create(Named("realvec") = "#", Named("realmat") = "##",
-      Named("boolvec") = "!", Named("boolmat") = "!!",
+    options = List::create(Named("realvec") = "##", Named("realmat") = "###",
+      Named("boolvec") = "!!", Named("boolmat") = "!!!",
       Named("charvec") = "$$", Named("charmat") = "$$$",
-      Named("intvec") = "%", Named("intmat") = "%%", 
+      Named("intvec") = "%%", Named("intmat") = "%%%", 
       Named("atomize") = false, Named("scalar") = true) ;
 
   RObject Expr = pl2r(A1, names, vars, options) ;
   RObject Res = Expr ;
   try
   {
-    Environment env = query_id->get_env() ;
-    Res = Language("dontCheck", Expr).eval(env) ;
+    Language id("dontCheck") ;
+    id.push_back(Expr) ;
+    Res = id.eval() ;
   }
 
   catch(std::exception& ex)
   {
-    throw PlException(PlCompound("r_eval", PlTermv(A1, PlAtom(ex.what())))) ;
-  }
-
-  catch(...)
-  {
-    throw PlException(PlCompound("r_eval", PlTermv(A1, PlString("exception occurred")))) ;
+    throw PlException(PlCompound("r_eval1", PlTermv(A1, PlTerm_atom(ex.what())))) ;
   }
 
   return true ;
@@ -1222,7 +1266,7 @@ PREDICATE(r_eval, 1)
 PREDICATE(r_eval, 2)
 {
   CharacterVector names ;
-  PlTerm vars ;
+  PlTerm_var vars ;
   List options ;
   if(query_id)
     options = query_id->get_options() ;
@@ -1232,27 +1276,33 @@ PREDICATE(r_eval, 2)
       Named("charvec") = "$$", Named("charmat") = "$$$",
       Named("intvec") = "%", Named("intmat") = "%%", 
       Named("atomize") = false, Named("scalar") = true) ;
-
+ 
   RObject Expr = pl2r(A1, names, vars, options) ;
   RObject Res = Expr ;
   try
   {
-    Environment env = query_id->get_env() ;
-    Res = Language("dontCheck", Expr).eval(env) ;
+    Language id("dontCheck") ;
+    id.push_back(Expr) ;
+    Res = id.eval() ;
   }
   
   catch(std::exception& ex)
   {
-    throw PlException(PlCompound("r_eval", PlTermv(A1, PlAtom(ex.what())))) ;
+    throw PlException(PlCompound("r_eval2", PlTermv(A1, PlTerm_atom(ex.what())))) ;
   }
 
-  catch(...)
+  PlTerm_var pl ;
+  try
   {
-    throw PlException(PlCompound("r_eval", PlTermv(A1, PlAtom("exception occurred")))) ;
+    PlCheckFail(pl.unify_term(r2pl(Res, names, vars, options))) ;
+  }
+  
+  catch(std::exception& ex)
+  {
+    throw PlException(PlCompound("r_eval2", PlTermv(A1, PlTerm_atom(ex.what())))) ;
   }
 
-  PlTerm pl = r2pl(Res, names, vars, options) ;
-  return A2 = pl ;
+  return A2.unify_term(pl) ;
 }
 
 // The SWI system should not be initialized twice; therefore, we keep track of
@@ -1299,3 +1349,146 @@ LogicalVector done_()
   pl_initialized = false ;
   return true ;
 }
+
+#endif // RPACKAGE
+ 
+#ifdef PROLOGPACK
+
+#include "SWI-cpp2.h"
+#include "RInside.h"
+
+RInside* r_instance = NULL ;
+
+PREDICATE(r_init, 0)
+{
+  if(r_instance)
+    return true ;
+
+  static int argc ;
+  static char** argv ;
+  if(!PL_is_initialised(&argc, &argv))
+  {
+    throw PlException(PlTerm_string("Prolog not initialized. Exiting.")) ;
+    return false ;
+  }
+
+  r_instance = new RInside(argc, argv) ;
+  return true ;
+}
+
+PREDICATE(r_eval_, 1)
+{
+  if(!R_TempDir)
+    throw PlException(PlTerm_string("R not initialized. Please invoke r_init.")) ;
+
+  CharacterVector names ;
+  PlTerm_var vars ;
+  List options ;
+  options = List::create(
+    Named("realvec") = "##", Named("realmat") = "###",
+    Named("boolvec") = "!!", Named("boolmat") = "!!!",
+    Named("charvec") = "$$", Named("charmat") = "$$$",
+    Named("intvec") = "%%", Named("intmat") = "%%%",
+    Named("atomize") = false, Named("scalar") = true) ;
+
+  RObject Expr = pl2r(A1, names, vars, options) ;
+  RObject Res = Expr ;
+  try
+  {
+    Language id("identity") ;
+    id.push_back(Expr) ;
+    Res = Rcpp_eval(id, Environment::global_env()) ;
+  }
+
+  catch(const Rcpp::eval_error& ex)
+  {
+    PlCompound syntax("evaluation_error", PlTermv(A1)) ;
+    PlCompound context("context", PlTermv(PlTerm_string("foreign r_eval_/2"), PlTerm_string(ex.what()))) ;
+    throw PlException(PlCompound("error", PlTermv(syntax, context))) ;
+  }
+  
+  catch(const std::exception& ex)
+  {
+    PlCompound syntax("evaluation_error", PlTermv(A1)) ;
+    PlCompound context("context", PlTermv(PlTerm_string("foreign r_eval_/2"), PlTerm_string(ex.what()))) ;
+    throw PlException(PlCompound("error", PlTermv(syntax, context))) ;
+  }
+
+  catch(...)
+  {
+    throw PlException(PlTerm_string("unknown exception")) ;
+    return false ;
+  }
+
+  return true ;
+}
+
+PREDICATE(r_eval_, 2)
+{
+  if(!R_TempDir)
+    throw PlException(PlTerm_string("R not initialized. Please invoke r_init.")) ;
+
+  CharacterVector names ;
+  PlTerm_var vars ;
+  List options ;
+  options = List::create(
+    Named("realvec") = "##", Named("realmat") = "###",
+    Named("boolvec") = "!!", Named("boolmat") = "!!!",
+    Named("charvec") = "$$", Named("charmat") = "$$$",
+    Named("intvec") = "%%", Named("intmat") = "%%%",
+    Named("atomize") = false, Named("scalar") = true) ;
+
+  RObject Expr = pl2r(A1, names, vars, options) ;
+  RObject Res = Expr ;
+  try
+  {
+    Language id("identity") ;
+    id.push_back(Expr) ;
+    Res = Rcpp_eval(id, Environment::global_env()) ;
+  }
+
+  catch(const Rcpp::eval_error& ex)
+  {
+    PlCompound syntax("evaluation_error", PlTermv(A1)) ;
+    PlCompound context("context", PlTermv(PlTerm_string("foreign r_eval_/2"), PlTerm_string(ex.what()))) ;
+    throw PlException(PlCompound("error", PlTermv(syntax, context))) ;
+  }
+
+  catch(const std::exception& ex)
+  {
+    PlCompound syntax("evaluation_error", PlTermv(A1)) ;
+    PlCompound context("context", PlTermv(PlTerm_string("foreign r_eval_/2"), PlTerm_string(ex.what()))) ;
+    throw PlException(PlCompound("error", PlTermv(syntax, context))) ;
+  }
+
+  catch(...)
+  {
+    throw PlException(PlTerm_string("unknown exception")) ;
+    return false ;
+  }
+
+  try
+  {
+    if(!A2.unify_term(r2pl(Res, names, vars, options)))
+    {
+      throw PlException(PlTerm_string("r_eval/2: Cannot unify R object.")) ;
+      return false ;
+    }
+  }
+
+  catch(std::exception& ex)
+  {
+    throw PlException(PlTerm_string(ex.what())) ;
+    return false ;
+  }
+
+  catch(...)
+  {
+    throw PlException(PlTerm_string("unknown exception1")) ;
+    return false ;
+  }
+
+  return true ;
+}
+
+#endif // PROLOGPACK
